#!/usr/bin/env python3

"""
Decimation algorithms to output low resolution bricks.
"""
##@package openzgy.impl.lodalgo
#@brief Decimation algorithms to output low resolution bricks.

import numpy as np
from enum import Enum
import warnings

from ..exception import ZgyUserError

##@brief Possible algorithms to generate LOD bricks.
class DecimationType(Enum):
    """
    Possible algorithms to generate LOD bricks.
    We might trim this list later to what is actually in use.
    The "classic" ZGY only uses the first two.

    CAVEAT: Many of these might be expensive to port and/or not
    possible to implement efficiently in Python.

    TODO-Low: Avoid exposing this enum to the public API.
    """
    LowPass = 0,           # Lowpass Z / decimate XY.
    WeightedAverage = 1,   # Weighted averaging (depends on global stats).
    Average = 2,           # Simple averaging.
    Median = 3,            # Somewhat more expensive averaging.
    Minimum = 4,           # Minimum value.
    Maximum = 5,           # Maximum value.
    MinMax = 6,            # Checkerboard of minimum and maximum values.
    Decimate = 7,          # Simple decimation, use first sample.
    DecimateSkipNaN = 8,   # Use first sample that is not NaN.
    DecimateRandom = 9,    # Random decimation using a fixed seed.
    AllZero = 10,          # Just fill the LOD brick with zeroes.
    WhiteNoise = 11,       # Fill with white noise, hope nobody notices.
    MostFrequent = 12,     # The value that occurs most frequently.
    MostFrequentNon0 = 13, # The non-zero value that occurs most frequently.
    AverageNon0 = 14,      # Average value, but treat 0 as NaN.

def _reorder(brick):
    """
    Reorder the input brick, decimating it by a factor 2 but keeping
    each of the 2x2x2 samples that are to be combined in the last
    dimension. E.g. an input shape (64, 64, 64) returns a (32, 32, 32, 8).
    """
    shape = (brick.shape[0]//2, brick.shape[1]//2, brick.shape[2]//2, 8)
    tmp = np.zeros(shape, dtype=brick.dtype)
    tmp[...,0] = brick[0::2,0::2,0::2]
    tmp[...,1] = brick[0::2,0::2,1::2]
    tmp[...,2] = brick[0::2,1::2,0::2]
    tmp[...,3] = brick[0::2,1::2,1::2]
    tmp[...,4] = brick[1::2,0::2,0::2]
    tmp[...,5] = brick[1::2,0::2,1::2]
    tmp[...,6] = brick[1::2,1::2,0::2]
    tmp[...,7] = brick[1::2,1::2,1::2]
    return tmp

_filter = np.array([
    +0.07996591908598821,
    -0.06968050331585399,
    -0.10596589191287473,
    +0.12716426995479813,
    +0.44963820678761110,
    +0.44963820678761110,
    +0.12716426995479813,
    -0.10596589191287473,
    -0.06968050331585399,
    +0.07996591908598821], dtype=np.float64) * 1.0392374478340252

def _decimate_LowPass(a):
    # TODO-Test, this needs very careful testing. I am sort of guessing here.
    # TODO-Medium, NaN and Inf handling compatible with what the C++ impl does.
    # TODO-Worry, if the algoritm doesn't have precisely zero DC response
    # I might need to explicitly test for all-constant traces to avoid
    # an empty fullres brick turning into a non-empty LOD brick.
    # See Salmon/Zgy/ArrayBasic/ArrayTile.cpp
    # See Salmon/Zgy/Common/Sampling.cpp
    # See numpy.convolve
    dtype = a.dtype
    a = a[::2,::2,:].astype(np.float32)
    a = np.pad(a, ((0,0), (0,0), (5,5)), mode = 'reflect')
    a = np.convolve(a.flatten('C'), _filter, mode = 'same').reshape(a.shape)
    a = a[..., 5:-5:2]
    return a.astype(dtype)

def _decimate_WeightedAverage(data, *, histogram, defaultvalue):
    """
    Decimate the input by weighted average, with the weights being the inverse
    of how common this sample value is in the entire survey. This means that
    the algorithm needs access to the survey histogram. Which might be a
    challenge if the histogram is being computed in parallel with the low
    resolution blocks.

    The computation can be run either in "storage" mode or in "float" mode.
    The data buffer and the histogram obviously need to be in the same mode.

    defaultvalue is used if all 8 inputs for one output sample are nan/inf.
    For float data this would almost always be 0.0 and for integral data
    the input samples cannot be nan/inf. So it might as well be 0 always.

    This fuction looks very different from the C++ version in the old ZGY
    accessor. The reason is that in C++ it is no big deal to have a tight
    inner loop processing one sample at a time. In Python that approach is
    a joke. On the other hand, Python has the fantastic numpy module.

    Note that I might choose to remove all NaN and Inf values in the
    _decimate_LowPass method that is typically used to generate lod 1.
    In that case the extra testung for NaN might be redundant.
    """

    if histogram is None or np.sum(histogram.bins) == 0:
        raise ZgyInternalError("The WeightedAverage algorithm needs a histogram.")

    # Linear transform from value to bin number.
    # Should round to nearest: bin = np.rint(value * factor + offset)
    r = histogram.vv_range
    n = histogram.bins.size
    factor = (n - 1)/ (r[1] - r[0])
    offset = -r[0] * factor

    # Better to be safe. All computation in double, to avoid surprises
    # e.g. with huge sample counts causing numeric underflow. Also handle
    # NaN / Inf values explicitly to avoid a lot of testing.
    input_dtype = data.dtype
    integral = np.issubdtype(input_dtype, np.integer)
    data = data.astype(np.float64, copy=True)
    if not integral:
        ugly = np.logical_not(np.isfinite(data))
        if np.any(ugly):
            data[ugly] = 0
        else:
            ugly = None
    else:
        ugly = None

    # Get histogram bin number for every input sample.
    bin_numbers = np.rint(data * factor + offset)
    np.clip(bin_numbers, 0, histogram.bins.size - 1, out = bin_numbers)
    bin_numbers = bin_numbers.astype(np.int32)

    # Get the sample's frequency for every input sample.
    src_frequency = histogram.bins[bin_numbers]

    # If frequency is reported as 0, this is an inconsistency.
    # Since we know it occurs at least one time.
    np.clip(src_frequency, 1, None, out=src_frequency)

    tmp_weight = np.float64(1) / src_frequency
    if ugly is not None:
        tmp_weight[ugly] = 0

    tmp_wsum = data * tmp_weight

    # Now sum each 2x2x2 set of samples.
    tmp_wsum = np.nansum(_reorder(tmp_wsum), 3)
    tmp_weight = np.nansum(_reorder(tmp_weight), 3)

    # Zero weights means no valid samples found among the 8 inputs.
    # This can only happen if non-finite data was present in the input,
    # causing weights to be set to zero earlier in this function.
    if ugly is not None:
        ugly2 = tmp_weight == 0
        tmp_wsum[ugly2] = defaultvalue
        tmp_weight[ugly2] = 1

    # Final result is the weighted sum of 8 input samples per output sample.
    # If all 8 samples were NaN we instead want to substitute an explicit
    # default value, or possibly set them back to to NaN.
    tmp_wsum /= tmp_weight

    # Convert back to user's type, protecting against integer overflow.
    if not integral:
        tmp_wsum = tmp_wsum.astype(np.float32, copy=False)
    else:
        cliprange = (np.iinfo(input_dtype).min, np.iinfo(input_dtype).max)
        np.clip(tmp_wsum, cliprange[0], cliprange[1], out=tmp_wsum)
        tmp_wsum = np.rint(tmp_wsum).astype(input_dtype, copy=False)

    return tmp_wsum

def _decimate_Average(brick):
    tmp = _reorder(brick.astype(np.double))
    # All NaN will return NaN, ditto for average of +inf, -inf.
    # I have not seen this documented explicitly but it should be
    # a safe assumption. Also the consequences if wrong are slight.
    # So, ignore the warning.
    with warnings.catch_warnings():
        warnings.simplefilter("ignore", category=RuntimeWarning)
        return np.nanmean(tmp, 3).astype(brick.dtype)

def _decimate_Median(brick):
    tmp = _reorder(brick.astype(np.double))
    with warnings.catch_warnings():
        warnings.simplefilter("ignore", category=RuntimeWarning)
        return np.nanmedian(tmp, 3).astype(brick.dtype)

def _decimate_Minimum(brick):
    tmp = _reorder(brick)
    with warnings.catch_warnings():
        warnings.simplefilter("ignore", category=RuntimeWarning)
        return np.nanmin(tmp, 3)

def _decimate_Maximum(brick):
    tmp = _reorder(brick)
    with warnings.catch_warnings():
        warnings.simplefilter("ignore", category=RuntimeWarning)
        return np.nanmax(tmp, 3)

def _decimate_Decimate(brick):
    return brick[0::2,0::2,0::2]

def _decimate_AllZero(brick):
    shape = (brick.shape[0]//2, brick.shape[1]//2, brick.shape[2]//2)
    return np.zeros(shape, dtype=brick.dtype)

def _decimate_AverageNon0(brick, *, defaultvalue = 0):
    """Average not considering 0, nan, inf. Use defaultvalue if all invalid."""
    tmp = _reorder(brick.astype(np.double))
    mask = np.isfinite(tmp)
    np.logical_not(mask, out=mask)
    np.logical_or(mask, tmp == 0, out=mask)
    tmp = np.ma.array(tmp, mask=mask, fill_value = defaultvalue)
    tmp = np.mean(tmp, 3)
    tmp.fill_value = defaultvalue # mean() doesn't preserve it
    return tmp.filled().astype(brick.dtype)

_decimation_algo = {
    DecimationType.LowPass:          _decimate_LowPass,
    DecimationType.WeightedAverage:  _decimate_WeightedAverage,
    DecimationType.Average:          _decimate_Average,
    DecimationType.Median:           _decimate_Median,
    DecimationType.Minimum:          _decimate_Minimum,
    DecimationType.Maximum:          _decimate_Maximum,
    #DecimationType.MinMax:           _decimate_MinMax,
    DecimationType.Decimate:         _decimate_Decimate,
    #DecimationType.DecimateSkipNaN:  _decimate_DecimateSkipNaN,
    #DecimationType.DecimateRandom:   _decimate_DecimateRandom,
    DecimationType.AllZero:          _decimate_AllZero,
    #DecimationType.WhiteNoise:       _decimate_WhiteNoise,
    #DecimationType.MostFrequent:     _decimate_MostFrequent,
    #DecimationType.MostFrequentNon0: _decimate_MostFrequentNon0,
    DecimationType.AverageNon0:      _decimate_AverageNon0,
}

def _is_power_of_two_ge_2(n):
    for shift in range(1, 32):
        if n == 1<<shift:
            return True
    return False

def _is_all_power_of_two_ge_2(nn):
    return np.all([_is_power_of_two_ge_2(n) for n in nn])

def _combine_eight_bricks(bricks):
    """
    Paste together 8 equal-sized bricks. Called only from decimate8(),
    which is not called from production code. Shoud perhaps be moved
    to the unit test module.
    """
    half = bricks[0].shape
    full = (half[0]*2, half[1]*2, half[2]*2)
    result = np.zeros(full, dtype=bricks[0].dtype)
    result[:half[0], :half[1], :half[2]] = bricks[0]
    result[:half[0], :half[1], half[2]:] = bricks[1]
    result[:half[0], half[1]:, :half[2]] = bricks[2]
    result[:half[0], half[1]:, half[2]:] = bricks[3]
    result[half[0]:, :half[1], :half[2]] = bricks[4]
    result[half[0]:, :half[1], half[2]:] = bricks[5]
    result[half[0]:, half[1]:, :half[2]] = bricks[6]
    result[half[0]:, half[1]:, half[2]:] = bricks[7]
    return result

def decimate(brick, algo, **kwargs):
    """
    Decimate a single input brick to produce one smaller output brick.
    """
    # The following test is only a requirement for bricksize.
    # decimate() can be called e.g. for entire brick columns,
    # and in that case the only requirement is that the sizes
    # are all even.
    #if not _is_all_power_of_two_ge_2(brick.shape):
    #    raise ZgyUserError("Brick size must be >= 2 and a power of 2")
    if not all([(n%2) == 0 for n in brick.shape]):
         raise ZgyUserError("Decimation can only be run on even-sized bricks")
    try:
        # TODO-Worry ... Remove the need for this kludge ...
        # TODO-Worry ... NOT THREADSAFE ... NOT THREADSAFE ... NOT THREADSAFE ...
        np_errors = np.seterr(all='print')
        if algo in _decimation_algo:
            return _decimation_algo[algo](brick, **kwargs)
        else:
            raise NotImplementedError(str(algo))
    finally:
        np.seterr(**np_errors)

def decimate8(bricks, algo, **kwargs):
    """
    Decimate 8 equal-sized input bricks to produce one output.

    Currently not used in production code. Only in the unit test.
    The genlod module prefers to split up and paste its data itself.
    This function shold perhaps be moved to the unit test module.

    Most of the decimation algorithms operate om a local 2x2x2 region
    for each output sample. So it makes no difference whether we
    decimate each input brick and then combine them, or the other
    way around. Combining last *might* use less memory if lazy
    evaluation works. Combining first might help the algorithm perform
    better (LowPass) or faster (WeightedAverage?)
    """
    test1 = decimate(_combine_eight_bricks(bricks), algo, **kwargs)
    test2 = _combine_eight_bricks([decimate(b, algo, **kwargs) for b in bricks])
    if algo != DecimationType.LowPass:
        assert np.all(test1 == test2)
    return test1

# Copyright 2017-2020, Schlumberger
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.



#!/usr/bin/env python3

"""
White-box unit tests for OpenZGY.
"""

import numpy as np
import os
import io
import sys

from ..impl import file as impl_file
from ..exception import *
from ..test.utils import SDCredentials, LocalFileAutoDelete, CloudFileAutoDelete, HasSeismicStore, SDTestData, SDTestSink

def checkThrows(e, msg, fn):
    try:
        fn()
        assert False and "Should have gotten an exception here."
    except e or ZgyUserError as ex:
        if msg:
            if str(ex).find(msg) < 0:
                print('Expect "{0}" got "{1}"'.format(msg, str(ex)))
            assert str(ex).find(msg) >= 0 and "Wrong exception message"
    except AssertionError:
        raise
    except Exception as ex:
        print('Expect "{0}" got "{1}"'.format(e or ZgyUserError, str(type(ex))))
        assert False and "Got the wrong type of exception."

class SimpleReceiver:
    def __call__(self, data, *args, **kwargs):
        self.data = bytes(data)

def _test_FileADT_read(f):
    """
    Check contents of a file created by test_SDFile.
    The test can be run both on a file descriptor still
    open for write and one recently opened read only.
    Also it works both for SD and local access.
    """
    # Simple interface
    data = f.xx_read(7, 9)
    assert data == b"continent"
    # Scatter/gather interface, a bit more roundabout to use.
    # Note: When reading back a SD file while it is still open
    # for write, the last (i.e. open) segment will be ' seg\n'.
    # This means that the part3 request spans the closed / open
    # segment boundary. Testing that corner case is important
    # because gthe ZGY library will encounter it very rarely and
    # only in conjunction with misaligned (compressed) files.
    part1 = SimpleReceiver()
    part2 = SimpleReceiver()
    part3 = SimpleReceiver()
    f.xx_readv([(1, 4, part1), (18, 7, part2), (60, 8, part3)])
    assert part1.data == b'ello'
    assert part2.data == b'this is'
    assert part3.data == b'last seg'
    assert f.xx_eof == 69
    checkThrows(None, "Invalid offset", lambda: f.xx_read(None, 1))
    checkThrows(None, "Invalid size", lambda: f.xx_read(0, None))
    checkThrows(None, "Invalid size", lambda: f.xx_read(0, 0))

def test_FileADT(filename):
    try:
        old_pread = os.pread if hasattr(os, "pread") else None
        with impl_file.FileFactory(filename, "w+b", "faux-context") as f:
            f.xx_write(b"hello, continent, THIS is a test.\n", 0)
            f.xx_write(b"hello again.\n", 34)
            f.xx_write(b"third block.\n", 47)
            f.xx_write(b"hello, continent, this is a test.\n", 0) # update
            f.xx_write(b"last sEG\n", 60)
            f.xx_write(b"eg", 66)
            checkThrows(None, "Invalid offset", lambda: f.xx_write("X", None))
            checkThrows(None, "Invalid offset", lambda: f.xx_write("X", -1))
            checkThrows(None, "Invalid size", lambda: f.xx_write(None, 0))
            checkThrows(None, "Invalid size", lambda: f.xx_write("", 0))
        with impl_file.FileFactory(filename, "rb", "faux-context") as f:
            assert f.threadsafe
            checkThrows(None, "open for writing", lambda: f.xx_write("X", 0))
            _test_FileADT_read(f)
        if old_pread: del os.pread
        with impl_file.FileFactory(filename, "rb", "faux-context") as f:
            assert not f.threadsafe
            _test_FileADT_read(f)
        checkThrows(None, "Opening ZGY as a is not supported.", lambda: impl_file.FileFactory(filename, "a", "faux-context"))
    finally:
        if old_pread: os.pread = old_pread

def test_SDFile(filename, buffered):
    with impl_file.FileFactory(filename, "w+b", SDCredentials(segsize=10/(1024*1024) if buffered else 0)) as f:
        assert not f.threadsafe
        assert f.xx_eof == 0
        f.xx_write(b"hello, continent, THIS is a test.\n", 0)
        f.xx_write(b"hello again.\n", 34)
        f.xx_write(b"third block.\n", 47)
        f.xx_write(b"hello, continent, this is a test.\n", 0) # update
        if not buffered:
            f.xx_write(b"last seg\n", 60)
        else:
            f.xx_write(b"last sEG\n", 60)
            f.xx_write(b"eg", 66) # writing to open segment
            # With segment size 10 and segment 0 unrestricted,
            # this is how the blocks should end up in storage.
            #   "hello, continent, THIS is a test.\n"
            #   "hello agai"
            #   "n.\nthird b"
            #   "lock.\nlast"
            #   " seg\n"
        assert f.xx_eof == 69
        checkThrows(ZgySegmentIsClosed, "write resized segment", lambda: f.xx_write(b"X", 34))
        checkThrows(ZgySegmentIsClosed, "write part of segment", lambda: f.xx_write(b"X", 35))
        checkThrows(ZgyUserError, "segments out of order", lambda: f.xx_write(b"X", 99))
        if not buffered:
            # When buffered this will be in the open segment, hence ok.
            checkThrows(ZgySegmentIsClosed, "write part of segment", lambda: f.xx_write(b"eg", 66))
            # The very ZGY specific limitation that all segments except
            # first and last must have the same size. Note that this will
            # not fail in the buffered test, because the buffering also
            # takes care of making all the segments the same size.
            checkThrows(ZgyUserError, "arbitrarily sized seg", lambda: f.xx_write(b"X", 69))
        # It should be possible to read back the file, even if some of it
        # is still sitting in the open segment. The test below also checks
        # reading across the closed / open segment boundary when buffered.
        _test_FileADT_read(f)
    with impl_file.FileFactory(filename, "rb", SDCredentials()) as f:
        assert f.threadsafe
        assert f.xx_eof == 69
        if not buffered:
            assert tuple(f._sizes) == (34, 13, 13, 9)
        else:
            assert tuple(f._sizes) == (34, 10, 10, 10, 5)
        _test_FileADT_read(f)
        # Test a read that crosses segment boundaries.
        slurp = f.xx_read(18, 45)
        assert slurp == b"this is a test.\nhello again.\nthird block.\nlas"
        # Test a read of a single, complete segment. There is a short cut.
        slurp = f.xx_read(34, 13)
        assert slurp == b"hello again.\n"
        # Test the readv interface in the same way.
        delivery = SimpleReceiver()
        f.xx_readv([(18, 45, delivery)])
        assert delivery.data == b"this is a test.\nhello again.\nthird block.\nlas"
        delivery = SimpleReceiver()
        f.xx_readv([(34, 13, delivery)])
        assert slurp == b"hello again.\n"

def test_SDFileReadPastEOF(filename):
    """
    Reading a block that crosses EOF should throw.
    """
    with impl_file.FileFactory(filename, "rb", SDCredentials()) as f:
        parts = list([SimpleReceiver() for i in range(8)])
        BS = 16 * 256 * 1024
        checkThrows(ZgyEndOfFile, None, lambda: (
            f.xx_readv([(0*BS, BS, parts[0]),
                        (1*BS, BS, parts[1]),
                        (2*BS, BS, parts[2]),
                        (3*BS, BS, parts[3])])
        ))
        checkThrows(ZgyEndOfFile, None, lambda: (
            f.xx_readv([(3*BS, BS, parts[7])])))

def test_Consolidate():
    """
    The logic for consolidating multiple requests into one is tricky
    enough to warrant a separate unit test.
    """
    class Want:
        def __init__(self, offset, size):
            if offset < 100000:
                # EOF is at 100000, we never get more than that.
                self._offset = offset
                self._size = min(offset + size, 100000) - offset
            else:
                # At or past EOF, we don't get any data at all.
                self._offset = None
                self._size = 0
            self.called = 0
        def __call__(self, data):
            self.called += 1
            data = np.frombuffer(data, dtype=np.uint8)
            if False:
                print("Delivery", (data[0], len(data)),
                      "expected", (self._offset, self._size),
                      "->", (self._offset % 251, self._size))
            assert self.called == 1
            assert len(data) == self._size
            if len(data):
                assert data[0] == self._offset % 251

    def test(requests, *args, **kwargs):
        # Add a functor that remembers what offset and size we ordered,
        # so that this can be checked when data gets delivered.
        requests = list([(offset, size, Want(offset, size)) for offset, size, _ in requests])
        # Invoke the function we are testing.
        result =  impl_file.FileADT._consolidate_requests(requests, *args, **kwargs)
        # Make a delivery of data where each element contains the absolute
        # offset of that element. Unless we pass our EOF marker.
        end_of_file = 100000
        for r in result:
            data = np.arange(r[0], min(r[0] + r[1], end_of_file), dtype=np.uint32)
            # Make it fit in a byte. Use a prime number modulus.
            data = (data % 251).astype(np.uint8)
            r[2](data.tobytes())
        # Check that each of our callbacks actually got called.
        for offset, size, fn in requests:
            assert fn.called == 1
        # Strip off the forward_result functor for easier checking of results.
        # Also convert to a tuple for the same reason.
        result = tuple([(e[0], e[1], None) for e in result])
        # Also strip functors off the input request, only for the below print()
        requests = tuple([(e[0], e[1], None) for e in requests])
        #print("in ", requests)
        #print("out", result)
        return result

    # Hole too large, no consolidation.
    requests = ((2, 98, None), (1000, 42, None))
    out = test(requests, max_hole = 300, max_size = None, force_align = None)
    assert out == requests

    # Same input, hole now deemed small enough.
    out = test(requests, max_hole = 1000, max_size = None, force_align = None)
    assert len(out) == 1
    assert out[0] == (2, 1040, None)

    # Same input, Hole too large, alignment was added.
    out = test(requests, max_hole = 300, max_size = None, force_align = 100)
    assert out == ((0, 100, None), (1000, 100, None))

    # Same input, hole now deemed small enough, alignment was added.
    out = test(requests, max_hole = 1000, max_size = None, force_align = 100)
    assert out[0] == (0, 1100, None)

    # max_size was exceeded.
    # Splitting is not very smart; see comments in source code.
    requests = list([(i, 70, None) for i in range(0, 66000, 100)])
    out = test(requests, max_hole = 300, max_size = 64000, force_align = None)
    assert out == ((0, 64000-30, None), (64000, 2000-30, None))

    # Requests larger than max_size
    requests = ((0, 1000, None), (1000, 2000, None), (3000, 500, None))
    out = test(requests, max_hole = 300, max_size = 10, force_align = None)
    assert out == requests

    # Request passes EOF
    requests = ((90000, 8000, None), # Inside
                (98000, 5000, None),  # Crosses EOF
                (103000, 5000, None), # Past EOF
    )
    out = test(requests, max_hole = 1000000, max_size = None, force_align = None)
    assert out == ((90000, 18000, None),)

    # Ditto but no consolidation, due to a small max_size.
    # No clipping needed in this case, as the lowest level
    # read will handle that.
    out = test(requests, max_hole = 300, max_size = 10, force_align = None)
    assert out == requests

if __name__ == "__main__":
    np.seterr(all='raise')
    with LocalFileAutoDelete("testfile.dat") as fn:
        test_FileADT(fn.name)
    if HasSeismicStore():
        with CloudFileAutoDelete("openzgy-1.dat", SDCredentials()) as fn:
            test_SDFile(fn.name, False)
        with CloudFileAutoDelete("openzgy-2.dat", SDCredentials()) as fn:
            test_SDFile(fn.name, True)
        test_SDFileReadPastEOF(SDTestData("Synt2.zgy"))
        test_Consolidate()
    sys.exit(0)

# Copyright 2017-2020, Schlumberger
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.



#!/usr/bin/env python3

"""
Efficient iterator for reading an entire ZGY file.

Naming conventions used in this file for parameters:

    surveysize: (i,j,k)-- Size of entire survey without padding.
    bricksize: (i,j,k) -- How data is stored. Usually (64, 64, 64).
    blocksize: (i,j,k) -- How much to read (at most) in one call.
    chunksize: (i,j,k) -- How much to return (at most) at once.
    maxbytes: int64    -- Max bytes to read at once, defaults to 1 GB
    dtype: np.dtype    -- Sample data type to return (float or storage).
    readfn: callable   -- To be called when reading a block.
    progress: callable -- Invoked after each read from disk.
"""
##@package openzgy.iterator
#@brief Efficient iterator for reading an entire ZGY file.

import numpy as np
from .api import _map_SampleDataTypeToDataType
from .impl.enum import _map_DataTypeToNumpyType

def _roundup(x, step):
    return ((x + step - 1) // step) * step

def _choose_blocksize(surveysize, bricksize, samplesize, maxbytes):
    """
    Get a reasonable blocksize (a.k.a. buffer size) to use when reading.
    Always read full traces, even if this means exceeding the specified
    maxbytes. Always read full bricks. Read more if it will fit inside
    the specified maxbytes.

    In each dimension the block size will either be the full survey size
    padded to bricksize or a power-of-two multiple of the brick size.
    In the simplest case caller can pass maxbytes=0 meaning read as little
    as possible; this will return (bricksize[0], bricksize[1], surveysize[2])
    a.k.a. one brick column.

    Note: It is desirable to have blocksize being a multiple of chunksize
    and chunksize being a multiple of bricksize. _choose_blocksize() tries
    to help with this: If there is room for more than one brick-column it
    will multiple the size in some direction, but only with a power of 2.
    This means that the adjustment will not break this condition.
    If it had not been for the power-of-two limitation this might not work.

    If the condition is met then the consumer knows that if a chunk smaller
    than the requested chunksize is received then this can only be due to
    reading at the end or bottom of the survey.
    """
    surveysize = list([_roundup(surveysize[i], bricksize[i]) for i in range(3)])
    bs = [bricksize[0], bricksize[1], surveysize[2]]
    while bs[0] * bs[1] * bs[2] * samplesize * 2 < maxbytes:
        bs[1] *= 2
    bs[1] = min(bs[1], surveysize[1])
    while bs[0] * bs[1] * bs[2] * samplesize * 2 < maxbytes:
        bs[0] *= 2
    bs[0] = min(bs[0], surveysize[0])
    return bs

def _readall_1(surveysize, blocksize, dtype, readfn):
    """
    Iterates over the entire file and returns data in blocks.
    This is the internal part of readall_2, it handles blocking the
    read but not chunking the data to be returned to the caller.
    """
    data  = np.zeros(blocksize, dtype=dtype) if readfn else None
    done  = np.int64(0)
    total = np.product(surveysize)

    for ii in range(0, surveysize[0], blocksize[0]):
        for jj in range(0, surveysize[1], blocksize[1]):
            for kk in range(0, surveysize[2], blocksize[2]):
                start = np.array((ii, jj, kk), dtype=np.int64)
                count = np.minimum(start + blocksize, surveysize) - start
                view = data[:count[0],:count[1],:count[2]] if data is not None else None
                #print("Reading", start, count, view.shape)
                if readfn: readfn(start, view)
                done += np.product(count)
                yield start, count, view
    assert done == np.product(surveysize)

def _readall_2(surveysize, blocksize, chunksize, dtype, readfn, progress):
    """
    Iterates over the entire file and returns data in chunks.
    All numeric and array parameters use numpy types.
    """
    done  = np.int64(0)
    total = np.product(surveysize)
    # Give a chance to abort before we even start.
    if progress and not progress(done, total): return
    alldata = _readall_1(surveysize=surveysize,
                         blocksize=blocksize,
                         dtype=dtype,
                         readfn=readfn)
    for datastart, datasize, data in alldata:
        for ii in range(0, datasize[0], chunksize[0]):
            for jj in range(0, datasize[1], chunksize[1]):
                # After reading a block but before yielding.
                if progress and not progress(done, total): return
                for kk in range(0, datasize[2], chunksize[2]):
                    start = np.array((ii, jj, kk), dtype=np.int64)
                    end   = np.minimum(start + chunksize, datasize)
                    count = end - start
                    view = data[start[0]:end[0],start[1]:end[1],start[2]:end[2]] if data is not None else None
                    done += np.product(count)
                    yield datastart + start, count, view
        # After yielding, give a chance to abort before the next read.
        # Also makes sure the final done==total is sent.
        if progress and not progress(done, total): return
    assert done == np.product(surveysize)

def _readall_3(surveysize, bricksize, blocksize, chunksize, dtype, readfn, maxbytes, progress):
    """
    Handle default arguments, make sure all data has numpy types,
    and then pass control to _readall_2.
    """
    surveysize = np.array(surveysize, dtype=np.int64)
    bricksize  = np.array(bricksize, dtype=np.int64)
    maxbytes   = np.int64(maxbytes if maxbytes is not None else 128*1024*1024)
    blocksize  = np.array(blocksize if blocksize is not None else _choose_blocksize(
        surveysize, bricksize, np.dtype(dtype).itemsize, maxbytes),
        dtype=np.int64)
    chunksize  = np.array(chunksize if chunksize is not None else blocksize, dtype=np.int64)
    blocksize[blocksize==0] = surveysize[blocksize==0]
    chunksize[chunksize==0] = blocksize[chunksize==0]
    if False:
        fmt = lambda x: "{0} = {1} voxels, {2:.1f} MB".format(str(tuple(x)), np.product(x), np.product(x) * dtype.itemsize/(1024*1024))
        print("survey", fmt(surveysize), "of", np.dtype(dtype).name)
        print("brick ", fmt(bricksize))
        print("block ", fmt(blocksize))
        print("chunk ", fmt(chunksize))
    # Since _readall_2 contains a yield, the function won't actually be
    # invoked until somebody consumes its output.
    result = _readall_2(surveysize=surveysize,
                      blocksize=blocksize,
                      chunksize=chunksize,
                      dtype=dtype,
                      readfn=readfn,
                      progress=progress)
    return result

def readall(r, *, blocksize = None, chunksize=None, maxbytes=128*1024*1024, dtype=None, cropsize = None, cropoffset = None, lod = 0, sizeonly = False, progress=None):
    """
    Convenience function to iterate over an entire cube, trying to use
    an efficient block size on read and optionally returning a smaller
    chunk size to the caller. The returned data buffer belongs to this
    iterator and may be overwritten on the next call. Two different
    iterators do not share their buffers.

    Note: If blocksize is a multiple of chunksize and chunksize is a
    multiple of bricksize then the following holds: if a chunk smaller
    than the requested chunksize is received then this can only be due
    to reading at the end or bottom of the survey. The caller might rely
    on this. E.g. when computing low resolution bricks and writing them
    out in a way that avoids a read/modify/write.

    CAVEAT: This function mingh be overkill in some cases.

    parameters:
        r: ZgyReader       -- The open ZGY file
        blocksize: (i,j,k) -- How much to read (at most) in one call.
                              If omitted, use a reasonable default.
                              Dims set to 0 mean "as much as possible".
        chunksize: (i,j,k) -- How much to return (at most) at once.
                              Defaults to blocksize, i.e. return the
                              same blocks that are read. Will never
                              return more than blocksize, so if set to
                              more than 64 you probably want to set an
                              explicit blocksize as well.
        maxbytes: int64    -- Max bytes to read at once, defaults to 128 MB
                              Ignored if an explicit blocksize is specified.
        dtype: np.dtype    -- Data type of returned buffer. Defaults to
                              what is in the file. Use np.float32 to
                              convert and scale (if needed) to float.
        cropsize: (i,j,k)  -- Only read this amount of data.
                              Note, no sanity check of this parameter.
        cropoffset: (i,j,k)   Start reading at this point.
                              Note, no sanity check of this parameter.
        lod: int              Pass >0 for decimated data.
        sizeonly: bool     -- If false, return sizes but not data.
        progress: callable -- Invoked after each read from disk.
    """
    if sizeonly:
        readfn = None
    elif cropoffset is None or tuple(cropoffset) == (0, 0, 0):
        readfn = lambda pos, data: r.read(pos, data, lod=lod)
    else:
        readfn = lambda pos, data: r.read(
            (pos[0]+cropoffset[0],
             pos[1]+cropoffset[1],
             pos[2]+cropoffset[2]),
            data, lod=lod)

    impl_dtype = _map_SampleDataTypeToDataType(r.datatype)
    dtype      = dtype or np.dtype(_map_DataTypeToNumpyType(impl_dtype))
    size       = np.array(r.size, dtype=np.int64)
    size       = (size + ((1<<lod) - 1)) // (1<<lod)
    return _readall_3(surveysize=cropsize if cropsize is not None else size,
                      bricksize=r.bricksize,
                      blocksize=blocksize,
                      chunksize=chunksize,
                      dtype=dtype,
                      readfn=readfn,
                      maxbytes=maxbytes,
                      progress=progress)

# Copyright 2017-2020, Schlumberger
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
